from goopylib.imports import *
from pathlib import Path as pathlib_Path

# I kinda wanted to scrap this, it wasn't that good.
def create_custom_ease():
    window = Window(title="goopylib: Create Custom Ease", width=get_screen_size()[1] * 0.7,
                    height=get_screen_size()[1] * 0.7, autoflush=False, bk_colour=DARKER_GREY)
    window.set_coords(0, 0, 1000, 1000)
    path = f"{pathlib_Path(__file__).parent.absolute()}/textures/"

    Image(Point(500, 500), f"{path}background.png").draw(window)

    add_button = Button(Image(Point(882, 219), f"{path}AddButton.png"),
                        Image(Point(882, 219), f"{path}AddButton.png").resize_factor(1.03),
                        Image(Point(882, 219), f"{path}AddButton.png").resize_factor(1.07),
                        Image(Point(882, 219), f"{path}AddButton.png").convert_greyscale()).draw(window)

    clear_button = Button(Image(Point(882, 280), f"{path}ClearButton.png"),
                          Image(Point(882, 280), f"{path}ClearButton.png").resize_factor(1.03),
                          Image(Point(882, 280), f"{path}ClearButton.png").resize_factor(1.07)).draw(window)

    play_button = Button(Image(Point(256, 805), f"{path}PlayButton.png"),
                         Image(Point(256, 805), f"{path}PlayButton.png").resize_factor(1.03),
                         Image(Point(256, 805), f"{path}PlayButton.png").resize_factor(1.07)).draw(window)

    shape_button = CycleButton(0,
                               Button(Image(Point(99, 805), f"{path}RectangleButton.png"),
                                      Image(Point(99, 805), f"{path}RectangleButton.png").resize_factor(1.03),
                                      Image(Point(99, 805), f"{path}RectangleButton.png").resize_factor(1.07)),

                               Button(Image(Point(99, 805), f"{path}CircleButton.png"),
                                      Image(Point(99, 805), f"{path}CircleButton.png").resize_factor(1.03),
                                      Image(Point(99, 805), f"{path}CircleButton.png").resize_factor(1.07))) \
        .draw(window)

    interpolation_button = CycleButton(0,
                                       Button(Image(Point(882, 109), f"{path}BezierButton.png"),
                                              Image(Point(882, 109), f"{path}BezierButton.png").resize_factor(1.03),
                                              Image(Point(882, 109), f"{path}BezierButton.png").resize_factor(1.07)),

                                       Button(Image(Point(882, 109), f"{path}CubicButton.png"),
                                              Image(Point(882, 109), f"{path}CubicButton.png").resize_factor(1.03),
                                              Image(Point(882, 109), f"{path}CubicButton.png").resize_factor(1.07)),

                                       Button(Image(Point(882, 109), f"{path}LinearButton.png"),
                                              Image(Point(882, 109), f"{path}LinearButton.png").resize_factor(1.03),
                                              Image(Point(882, 109), f"{path}LinearButton.png").resize_factor(1.07))) \
        .draw(window)

    template_button = CycleButton(0,
                                  Button(Image(Point(882, 411), f"{path}LinearTemplate.png"),
                                         Image(Point(882, 411), f"{path}LinearTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}LinearTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}BackTemplate.png"),
                                         Image(Point(882, 411), f"{path}BackTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}BackTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}ExponentialTemplate.png"),
                                         Image(Point(882, 411), f"{path}ExponentialTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}ExponentialTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}PolynomialTemplate.png"),
                                         Image(Point(882, 411), f"{path}PolynomialTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}PolynomialTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}SineTemplate.png"),
                                         Image(Point(882, 411), f"{path}SineTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}SineTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}CircleTemplate.png"),
                                         Image(Point(882, 411), f"{path}CircleTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}CircleTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}ElasticTemplate.png"),
                                         Image(Point(882, 411), f"{path}ElasticTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}ElasticTemplate.png").resize_factor(1.07)),

                                  Button(Image(Point(882, 411), f"{path}BounceTemplate.png"),
                                         Image(Point(882, 411), f"{path}BounceTemplate.png").resize_factor(1.03),
                                         Image(Point(882, 411), f"{path}BounceTemplate.png").resize_factor(1.07)),

                                  disabled_graphic=Image(Point(882, 428), f"{path}CustomTemplate.png")).draw(window)

    save_button = Button(Image(Point(882, 647), f"{path}SaveButton.png"),
                         Image(Point(882, 647), f"{path}SaveButton.png").resize_factor(1.03),
                         Image(Point(882, 647), f"{path}SaveButton.png").resize_factor(1.07)).draw(window)

    open_button = Button(Image(Point(882, 708), f"{path}OpenButton.png"),
                         Image(Point(882, 708), f"{path}OpenButton.png").resize_factor(1.03),
                         Image(Point(882, 708), f"{path}OpenButton.png").resize_factor(1.07)).draw(window)

    simulation_graphic1 = CycleButton(0, Rectangle(Point(50, 875), Point(125, 950), fill=LIGHTER_BLUE, outline_width=0),
                                      Circle(Point(88, 913), 38, fill=LIGHTER_BLUE, outline_width=0),
                                      autoflush=False).draw(window)

    simulation_graphic2 = CycleButton(0,
                                      Rectangle(Point(845, 845), Point(920, 920), fill=LIGHTER_BLUE, outline_width=0),
                                      Circle(Point(883, 883), 37.5, fill=LIGHTER_BLUE, outline_width=0),
                                      autoflush=False).draw(window)

    graph = Image(Point(500, 500), f"{path}Graph.png").draw(window)
    delete_dropdown = Button(Image(Point(0, 0), f"{path}DeleteDropdown.png"),
                             Image(Point(0, 0), f"{path}DeleteDropdownHover.png"))

    colour_grad = colour_gradient(LIGHTER_VIOLET, DARKEST_VIOLET, 100)

    resolution = 2

    control_points = [Circle(p, radius=10, fill=VIOLET, outline_width=0).draw(window) for p in
                      [Point(110, 673), Point(668, 118)]]
    last_control_points = control_points.copy()

    points = []
    for obj in control_points:
        points.append(obj.anchor)

    curve = []
    for t in range(int(10 ** resolution)):
        t /= 10 ** resolution
        curve.append(py_bezier_curve(t, points))

    circle_objects = []
    for i in range(int(10 ** resolution)):
        circle_objects.append(Circle(curve[i], radius=2, fill=colour_grad[i], outline_width=0).draw(window))

    selected_point = None
    selected_curve_point = None

    def ease(time):
        points = []
        for obj in control_points:
            points.append(obj.anchor)

        if interpolation_button.get_state() == 0:
            return 1 - (py_bezier_curve(time, points).y - 118) / 555
        else:
            return 1 - (LinearInterpolation(time, points).y - 118) / 555

    def update_curve():
        nonlocal circle_objects, curve, point, last_control_points, t, i

        points = []
        for obj in control_points:
            points.append(obj.anchor)

        curve = []
        for t in range(int(10 ** resolution)):
            t /= 10 ** resolution

            if interpolation_button.get_state() == 0:
                curve.append(py_bezier_curve(t, points))
            else:
                curve.append(LinearInterpolation(t, points))

        for i in range(int(10 ** resolution)):
            circle_objects[i].move_to_y(curve[i].y)

        last_control_points = control_points.copy()

    while True:
        t = time.time()
        mouse_pos = window.check_left_mouse_click()
        if mouse_pos is not None:
            if open_button.is_clicked(mouse_pos):
                filename = openfilebrowser()

            elif save_button.is_clicked(mouse_pos):
                filename = openfilebrowser()

            elif interpolation_button.is_clicked(mouse_pos):
                update_curve()

            elif play_button.is_clicked(mouse_pos):
                play_button.disable()

                """
                x = []
                for t in range(int(10 ** resolution)):
                    t /= 10 ** resolution
                    x.append(ease(t))
                plt.plot(x)
                plt.show()"""

                simulation_graphic1.glide_x(500, time=2, easing=ease)
                simulation_graphic2.animate_rotate(360, time=2, easing=ease)

            elif shape_button.is_clicked(mouse_pos):
                simulation_graphic1.set_state(shape_button.get_state())
                simulation_graphic2.set_state(shape_button.get_state())

            elif clear_button.is_clicked(mouse_pos):
                for point in control_points[1:-1]:
                    point.undraw()
                    control_points.remove(point)

            elif add_button.is_clicked(mouse_pos):
                add_button.disable()
                if selected_curve_point is None:
                    i = int((10 ** resolution / 2) * 2 ** (2 - len(control_points)))
                    control_points.insert(1, Circle(curve[i], radius=10, fill=colour_grad[i], outline_width=0).draw(
                        window))
                    control_points[1].set_draggable(callback=update_curve)
                else:
                    i = max(
                        math.ceil(circle_objects.index(selected_curve_point) / (100 / (len(control_points) - 1))) - 1,
                        1)
                    control_points.insert(i, Circle(selected_curve_point.get_anchor(), radius=10,
                                                    fill=colour_grad[circle_objects.index(selected_curve_point)],
                                                    outline_width=0).draw(window))
                    selected_curve_point = None
                    control_points[i].set_draggable(callback=update_curve)

                add_button.enable()

                template_button.disable()

            elif delete_dropdown.is_clicked(mouse_pos):
                delete_dropdown.undraw()
                selected_point.undraw()
                control_points.remove(selected_point)

                if len(control_points) == 2:
                    template_button.enable()

            else:
                for point in circle_objects:
                    if point.is_clicked(mouse_pos):
                        for i, p in enumerate(circle_objects):
                            p.set_fill(colour_grad[i])
                        point.set_fill(WHITE)
                        selected_curve_point = point
                        break

        if last_control_points != control_points:
            update_curve()

        if play_button.is_disabled and not simulation_graphic1.is_gliding:
            play_button.enable()

        mouse_pos = window.check_right_mouse_click()
        for point in control_points[1:-1]:
            if point.is_clicked(mouse_pos):
                delete_dropdown.draw(window).move_to_point(point.anchor, align="topleft")
                selected_point = point

        window.update()


def create_custom_ease2():
    window = Window(title="goopylib_b: Create Custom Ease", width=get_screen_size()[1] * 0.7,
                    height=get_screen_size()[1] * 0.7, autoflush=False, bk_colour=DARKER_GREY)
    window.set_coords(0, 0, 1000, 1000)
    path = f"{pathlib_Path(__file__).parent.absolute()}/textures/"

    while True:
        if window.is_closed():
            break

        window.update()

    window.close()



from goopylib.colours import *
from goopylib.util import GraphicsError

global_style = "default"

STYLES = {"default": {"fill": CHROME_YELLOW, "background": WHITE,
                      "outline": BLACK, "line outline": GREY, "width": 2,

                      # Config for the Line Class
                      "line width": 3, "capstyle": "round", "joinstyle": "round", "arrow": "none",
                      "arrowshape": (8, 10, 3), "dash": None,

                      # Config for the Entry class
                      "entry width": 0, "select colour": BLUE,

                      # Config for Font & Text
                      "text": "Lorem Ipsum", "justify": "center", "font face": "calibri", "text align": "center",
                      "font size": 10, "font style": "normal", "font colour": BLACK},
          }


# Add other colour definitions here!

############################################################################
# Graphics classes start here

def set_style(style="default"):
    global global_style

    if style not in STYLES.keys():
        raise GraphicsError(f"The style specified ({style}) does not exist, must be one of {list(STYLES.keys())}")
    global_style = style



from goopylib.objects.CycleButton import CycleButton
from goopylib.util import GraphicsError

class Checkbox(CycleButton):
    def __init__(self, true_graphic, false_graphic, disabled_graphic=None, autoflush=True, state=True, layer=0,
                 tag=None, bounds=None):

        super().__init__(true_graphic, false_graphic, disabled_graphic=disabled_graphic, state=int(not state),
                         layer=layer, autoflush=autoflush, tag=tag, bounds=bounds)
        self.bound_objects = {*()}

    def __repr__(self):
        return f"Checkbox({self.states[0]}, {self.states[1]})"

    def set_state(self, state):
        super().set_state(not int(state))
        for obj in self.bound_objects:
            obj.set_state(state)
        return self

    def get_state(self):
        return bool(self.state)

    def change_state(self):
        self.set_state(not self.state)
        return self

    def bind_state_to(self, other):
        if not isinstance(other, Checkbox):
            raise GraphicsError(f"Object to bind this Checkbox to must be another Checkbox, not {other}")
        other.bound_objects.add(self)

        self.autoflush = False
        self.set_state(other.state)

        return self



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.styles import *

class Text(GraphicsObject):

    def __init__(self, p, text, font_size=None, font_face=None, font_style=None, font_colour=None,
                 justify="left", text_align="center", layer=0, tag=None, bounds=None):

        self.anchor = p.copy()
        GraphicsObject.__init__(self, layer=layer, tag=tag, bounds=bounds)

        if isinstance(font_colour, Colour):
            self.outline = font_colour
        elif font_colour is None:
            self.outline = STYLES["default"]["outline"]
        else:
            raise GraphicsError("\n\nGraphicsError: font colour for Text object must be a Colour type object, "
                                f"not {font_colour}")

        if isinstance(font_face, str):
            self.font_face = font_face.lower()
        elif font_face is None:
            self.font_face = STYLES["default"]["font face"]
        else:
            raise GraphicsError(f"\n\nGraphicsError: font face for Text object must be a string, not {font_colour}")

        if isinstance(font_style, str):
            self.font_style = font_style.lower()
        elif font_style is None:
            self.font_style = STYLES["default"]["font style"]
        else:
            raise GraphicsError(f"\n\nGraphicsError: font style for Text object must be a string, not {font_style}")

        if isinstance(text_align, str):
            self.text_align = text_align.lower()
            if self.text_align == "top":
                self.text_align = "n"
            elif self.text_align == "topright":
                self.text_align = "ne"
            elif self.text_align == "right":
                self.text_align = "e"
            elif self.text_align == "bottomright":
                self.text_align = "se"
            elif self.text_align == "bottom":
                self.text_align = "s"
            elif self.text_align == "bottomleft":
                self.text_align = "sw"
            elif self.text_align == "left":
                self.text_align = "w"
            elif self.text_align == "topleft":
                self.text_align = "nw"
            elif self.text_align == "center":
                pass
            else:
                raise GraphicsError(
                    "Text align must be one one of ['top', 'topright', 'right', 'bottomright', 'bottom', "
                    f"'bottomleft', 'left', 'topleft', 'center'], not {text_align}")
        elif text_align is None:
            self.text_align = STYLES["default"]["text align"]
        else:
            raise GraphicsError(f"\n\nGraphicsError: text align for Text object must be a string, not {text_align}")

        if isinstance(justify, str):
            self.justify = justify.lower()
        elif justify is None:
            self.justify = STYLES["default"]["justify"]
        else:
            raise GraphicsError(f"\n\nGraphicsError: justify for Text object must be a string, not {justify}")

        if isinstance(font_size, int):
            self.font_size = font_size
        elif font_size is None:
            self.font_size = STYLES["default"]["font size"]
        else:
            raise GraphicsError(f"\n\nGraphicsError: font size for Text object must be an integer, not {font_size}")

        self._initial_font_size = self.font_size

        self.set_text_colour = self.set_fill

        self.text = str(text)

    def __repr__(self):
        return f"Text({self.anchor}, '{self.text}')"

    def _draw(self, canvas, options):
        x, y = canvas.to_screen(self.anchor[0], self.anchor[1])

        self.set_size(abs(int(self._initial_font_size / canvas.trans.x_scale)), False)

        return canvas.create_text(x, y, font=(self.font_face, self.font_size, self.font_style),
                                  anchor=self.text_align, justify=self.justify, text=self.text, fill=self.outline)

    def _move(self, dx, dy):
        self.anchor[0] += dx
        self.anchor[1] += dy

    def is_clicked(self, mouse_pos):
        if mouse_pos is None:
            return False
        if self.bounds is None:
            width = self.get_width()
            height = self.get_height()

            if self.text_align == "center":
                anchor = self.anchor
            if self.text_align == "n":
                anchor = [self.anchor[0], self.anchor[1] + height // 2]
            elif self.text_align == "ne":
                anchor = [self.anchor[0] - width // 2, self.anchor[1] + height // 2]
            elif self.text_align == "e":
                anchor = [self.anchor[0] - width // 2, self.anchor[1]]
            elif self.text_align == "se":
                anchor = [self.anchor[0] - width // 2, self.anchor[1] - height // 2]
            elif self.text_align == "s":
                anchor = [self.anchor[0], self.anchor[1] - height // 2]
            elif self.text_align == "sw":
                anchor = [self.anchor[0] + width // 2, self.anchor[1] - height // 2]
            elif self.text_align == "w":
                anchor = [self.anchor[0] + width // 2, self.anchor[1]]
            elif self.text_align == "nw":
                anchor = [self.anchor[0] + width // 2, self.anchor[1] + height // 2]

            if (anchor[0] - width // 2 < mouse_pos[0] < anchor[0] + width // 2) and \
                    (anchor[1] - height // 2 < mouse_pos[1] < anchor[1] + height // 2):
                return True
            return False

        return self.bounds.is_clicked(mouse_pos)

    def get_width(self):
        return self.font_size * len(self.text)

    def get_height(self):
        return self.font_size

    def clone(self, new_tag=None):
        other = Text(self.anchor, self.text, font_size=self._initial_font_size, font_face=self.font_face,
                     font_style=self.font_style, font_colour=self.outline, justify=self.justify,
                     text_align=self.text_align, layer=self.layer, tag=new_tag, bounds=self.bounds.clone())
        return other

    def set_text(self, text):
        self.text = str(text)
        self._update_layer()
        return self

    def get_text(self):
        return self.text

    def get_anchor(self):
        return self.anchor.copy()

    def set_face(self, face):
        if isinstance(face, str):
            self.font_face = face
        else:
            raise GraphicsError(f"\n\nGraphicsError: The font face for the text must be a string, not {face}")
        self._update_layer()
        return self

    def set_text_align(self, text_align):
        if isinstance(text_align, str):
            self.text_align = text_align
        else:
            raise GraphicsError(f"\n\nGraphicsError: The text_align for the text must be a string, not {text_align}")

        if self.text_align == "top":
            self.text_align = "n"
        elif self.text_align == "topright":
            self.text_align = "ne"
        elif self.text_align == "right":
            self.text_align = "e"
        elif self.text_align == "bottomright":
            self.text_align = "se"
        elif self.text_align == "bottom":
            self.text_align = "s"
        elif self.text_align == "bottomleft":
            self.text_align = "sw"
        elif self.text_align == "left":
            self.text_align = "w"
        elif self.text_align == "topleft":
            self.text_align = "nw"
        elif self.text_align == "center":
            pass
        else:
            raise GraphicsError("\n\nGraphicsError: Text align must be one one of ['top', 'topright', 'right', "
                                f"'bottomright', 'bottom', 'bottomleft', 'left', 'topleft', 'center'] not {text_align}")
        self._update_layer()
        return self
    
    def set_justify(self, justify):
        if isinstance(justify, str):
            self.justify = justify
        else:
            raise GraphicsError("\n\nGraphicsError: Text justify must be one one of ['top', 'topright', 'right', "
                                f"'bottomright', 'bottom', 'bottomleft', 'left', 'topleft', 'center'], not {justify}")
        self._update_layer()
        return self

    def set_size(self, size, _update_init=True):
        if isinstance(size, int):
            self.font_size = size
        else:
            raise GraphicsError(f"\n\nGraphicsError: The size for the text must be an integer, not {size}")

        if _update_init:
            self._initial_font_size = self.font_size

        self._update_layer()
        return self
        
    def set_style(self, style):
        if isinstance(style, str):
            if style in {'bold', 'normal', 'italic', 'bold italic'}:
                self.font_style = style
            else:
                raise GraphicsError("Text style must be one of ['bold', 'normal', 'italic', 'bold italic']")
        else:
            raise GraphicsError("\n\nGraphicsError: Text style must be one of ['bold', 'normal', 'italic', "
                                f"'bold italic'], not {style}")

        self._update_layer()
        return self



import goopylib.objects._BBox as gpBBox
from goopylib.util import GraphicsError


class Rectangle(gpBBox.BBox):
    def __init__(self, p1, p2, bounds=None, fill=None, outline=None, outline_width=None, layer=0,
                 cursor="arrow", is_rounded=False, roundness=5, tag=None):

        # This makes sure that the p1 & p2 arguments of the Rectangle are lists, raises an error otherwise
        if not isinstance(p1, list):
            raise GraphicsError("\n\nGraphicsError: p1 argument for Rectangle must be a list in the form [x, y], "
                                f"not {p1}")
        if not isinstance(p2, list):
            raise GraphicsError("\n\nGraphicsError: p2 argument for Rectangle must be a list in the form [x, y], "
                                f"not {p2}")

        # Makes sure that the roundness argument is an integer, raises an error otherwise
        if not isinstance(roundness, int):
            raise GraphicsError("\n\nGraphicsError: roundness argument for Rectangle must be an integer, "
                                f"not {roundness}")

        # A call to the super class to initialize important variables of the Rectangle class.
        gpBBox.BBox.__init__(self, p1, p2, bounds=bounds, fill=fill, outline=outline, outline_width=outline_width,
                             cursor=cursor, layer=layer, tag=tag)

        self.is_rounded = is_rounded  # A variable defining whether the rectangle is rounded
        self.sharpness = roundness  # Usually values between 2 & 10 work well.

    def __repr__(self):
        return f"Rectangle({self.p1}, {self.p2}, tag={self.tag})"

    def _set_resizable(self, resizables, top_bounds=None, bottom_bounds=None, left_bounds=None, right_bounds=None,
                       thickness=10):

        # resizables is a list [top, bottom, left, right] where each of its elements are Booleans defining whether the
        # rectangle is resizable in that direction

        self.resizing_bounds = {}  # Stores the bounds of the resizing objects for 'top', 'bottom', 'left', & 'right'
        if resizables[0]:  # Checks if the object is set to be resized from the top
            if top_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["top"] = Rectangle([self.p1[0] - thickness / 2, self.p1[1] - thickness / 2],
                                                        [self.p2[0] + thickness / 2, self.p1[1] + thickness / 2],
                                                        cursor="vertical-resizing-arrow")
            elif isinstance(top_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["top"] = top_bounds

            elif top_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["top"] = GraphicsObject.tagged_objects[top_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

        if resizables[1]:  # Checks if the object is set to be resized from the bottom
            if bottom_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["bottom"] = Rectangle([self.p2[0] + thickness / 2, self.p2[1] - thickness / 2],
                                                           [self.p1[0] - thickness / 2, self.p2[1] + thickness / 2],
                                                           cursor="vertical-resizing-arrow")
            elif isinstance(bottom_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["bottom"] = top_bounds

            elif bottom_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["bottom"] = GraphicsObject.tagged_objects[top_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

        if resizables[2]:  # Checks if the object is set to be resized from the left
            if left_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["left"] = Rectangle([self.p1[0] - thickness / 2, self.p1[1] - thickness / 2],
                                                         [self.p1[0] + thickness / 2, self.p2[1] + thickness / 2],
                                                         cursor="horizontal-resizing-arrow")
            elif isinstance(left_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["left"] = left_bounds

            elif left_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["left"] = GraphicsObject.tagged_objects[left_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

        if resizables[3]:  # Checks if the object is set to be resized from the right
            if right_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["right"] = Rectangle([self.p2[0] + thickness / 2, self.p2[1] + thickness / 2],
                                                          [self.p2[0] - thickness / 2, self.p1[1] - thickness / 2],
                                                          cursor="horizontal-resizing-arrow")
            elif isinstance(right_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["right"] = right_bounds

            elif right_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["right"] = GraphicsObject.tagged_objects[right_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

    def _draw(self, canvas, options):
        if self.is_rounded:  # The rectangle is rounded, so we draw a polygon
            points = [self.p1[0], self.p1[1], self.p1[0], self.p2[1], self.p2[0], self.p2[1], self.p2[0], self.p1[1]]
            for point in range(len(points) - 2):
                points[point * 2], points[point * 2 + 1] = canvas.to_screen(points[point * 2], points[point * 2 + 1])

            # Code modified from Francisco Gomes, https://stackoverflow.com/users/9139005/francisco-gomes

            # The sharpness here is just how close the sub-points
            # are going to be to the vertex. The more the sharpness,
            # the more the sub-points will be closer to the vertex.
            # (This is not normalized)

            x = points[::2]
            y = points[1::2]

            if self.sharpness < 2:
                self.sharpness = 2

            ratio_multiplier = self.sharpness - 1
            ratio_dividend = self.sharpness

            # List to store the points
            points = []

            # Iterate over the x points
            for i in range(len(x)):
                # Set vertex
                points.append(x[i])
                points.append(y[i])

                # If it's not the last point
                if i != (len(x) - 1):
                    # Insert submultiples points. The higher the sharpness, the more these points will be
                    # closer to the vertex.
                    points.append((ratio_multiplier * x[i] + x[i + 1]) / ratio_dividend)
                    points.append((ratio_multiplier * y[i] + y[i + 1]) / ratio_dividend)
                    points.append((ratio_multiplier * x[i + 1] + x[i]) / ratio_dividend)
                    points.append((ratio_multiplier * y[i + 1] + y[i]) / ratio_dividend)
                else:
                    # Insert submultiples points.
                    points.append((ratio_multiplier * x[i] + x[0]) / ratio_dividend)
                    points.append((ratio_multiplier * y[i] + y[0]) / ratio_dividend)
                    points.append((ratio_multiplier * x[0] + x[i]) / ratio_dividend)
                    points.append((ratio_multiplier * y[0] + y[i]) / ratio_dividend)
                    # Close the polygon
                    points.append(x[0])
                    points.append(y[0])

            # This is done due to an internal bug in Tkinter where it does not set the width of the polygon...
            if self.outline_width == 0:
                self.outline = self.fill

            return canvas.create_polygon(points, width=self.outline_width, fill=self.fill, outline=self.outline,
                                         smooth=self.is_rounded)

        else:  # The rectangle is not rounded, so we draw a rectangle to the window
            x1, y1 = canvas.to_screen(self.p1[0], self.p1[1])  # Transforms the 2 points to match with the scaling of
            x2, y2 = canvas.to_screen(self.p2[0], self.p2[1])  # the Window

            return canvas.create_rectangle(x1, y1, x2, y2, fill=self.fill,  # actually creates the rectangle
                                           outline=self.outline, width=self.outline_width)

    def clone(self, new_tag=None):
        other = Rectangle(self.p1, self.p2, fill=self.fill, outline=self.outline, outline_width=self.outline_width,
                          layer=self.layer, cursor=self.cursor, is_rounded=self.is_rounded, roundness=self.sharpness,
                          tag=new_tag, bounds=self.bounds.clone())
        return other

    def is_clicked(self, pos):
        if pos is None:
            return False

        if self.bounds is None:
            width = self.width // 2
            height = self.height // 2
            return self.anchor[0] - width < pos[0] < self.anchor[0] + width and \
                   self.anchor[1] - height < pos[1] < self.anchor[1] + height
        else:
            return self.bounds.is_clicked(pos)


from goopylib.objects.GraphicsObject import GraphicsObject



from goopylib.objects.CurvedLine import CurvedLine
from goopylib.util import GraphicsError

import math

class Arc(CurvedLine):
    def __init__(self, p, start_angle, end_angle, radius, radius2=None, outline=None, outline_width=None,
                 cursor="arrow", arrow=None, resolution=10, smooth=True, bounds_width=10, layer=0, tag=None):

        if not isinstance(p, list):
            raise GraphicsError(f"\n\nGraphicsError: anchor for arc (p) must be a list in the form [x, y], not {p}")
        
        if not (isinstance(start_angle, int) or isinstance(start_angle, float)):
            raise GraphicsError(f"\n\nGraphicsError: start_angle must be an integer or float, not {start_angle}")
        if not (isinstance(end_angle, int) or isinstance(end_angle, float)):
            raise GraphicsError(f"\n\nGraphicsError: end_angle must be an integer or float, not {end_angle}")
        
        if not (isinstance(radius, int) or isinstance(radius, float)):
            raise GraphicsError(f"\n\nGraphicsError: Arc radius must be an integer or float, not {radius}")
        if radius2 is not None:
            if not (isinstance(radius2, int) or isinstance(radius2, float)):
                raise GraphicsError(f"\n\nGraphicsError: Arc radius2 must be an integer or float, not {radius2}")
        
        self.start_angle = start_angle
        self.end_angle = end_angle

        self.anchor = p

        self.radius1 = radius
        if radius2 is None:
            self.radius2 = radius
        else:
            self.radius2 = radius

        angle_change = (end_angle - start_angle) / resolution

        range_end = 90 / angle_change
        decimal_points = min([len(str(range_end).split('.')[1]), 3])

        x_coeff = sum([abs(math.cos(math.radians(i * angle_change + start_angle)))
                       for i in range(int(range_end * 10 ** decimal_points))])
        y_coeff = sum([abs(math.sin(math.radians(i * angle_change + start_angle)))
                       for i in range(int(range_end * 10 ** decimal_points))])

        x_coeff /= 10 ** decimal_points
        y_coeff /= 10 ** decimal_points

        angle_change = (end_angle - start_angle) / resolution

        x_change = self.radius1 / 2 / x_coeff
        y_change = self.radius2 / 2 / y_coeff

        self.points = [[p[0] - x_change/2, p[1] - self.radius2 / 2]]

        for i in range(resolution):
            cur_angle = (i * angle_change) + start_angle
            self.points.append([self.points[-1][0] + x_change * math.cos(math.radians(cur_angle)), self.points[-1][1] + y_change * math.sin(math.radians(cur_angle))])

        CurvedLine.__init__(self, *self.points, outline=outline, outline_width=outline_width, arrow=arrow,
                            resolution=0, interpolation="spline", smooth=smooth, bounds_width=bounds_width, layer=layer,
                            tag=tag)



from ctypes import c_buffer, windll
from time import sleep
from sys import getfilesystemencoding

from goopylib.util import GraphicsError

# Right now, the Sound module is only compatible with windows & wav files

# Support needed for wav, mp3, m4a, flac, wma, aac, MIDI, MusicXML

"""
ctypes is a library that implements C data types and allows you to use DLLs 

https://support.microsoft.com/en-in/help/815065/what-is-a-dll
DLLs (Digital Link Library) are libraries written (by?) for Microsoft that can control general elements of windows 
    .cpl are Control Panel DLLs and can control Control Panel Setings
    .ocx are ActiveXControls. For example, the Condlg32 DLL allows you to create dialog boxes 
    .drv are Device Drivers. For example, they can control printing to a printer. 

MCIs are basically multimedia devices and they can be controlled using certain commands.

Open Device -> Send Commands -> Close Device 

MCI Commands:

    open: https://docs.microsoft.com/en-us/windows/win32/multimedia/open
    close: https://docs.microsoft.com/en-us/windows/win32/multimedia/close

    ----------------------------------------------------------------------------------------------------
    set: https://docs.microsoft.com/en-us/windows/win32/multimedia/set	
    SET FLAGS FOR WAVEAUDIO:

        "set {alias} {flags}"

        alignment {integer} - Sets the alignment of data blocks relative to the start of data passed to the 
                              waveform-audio device.

                              "Data alignment refers to where data is located in memory. All data is at least 
                              1-byte aligned, meaning that it starts at any one byte in memory. Data that is 
                              n-byte aligned is located somewhere with a memory address that is an exact multiple 
                              of n bytes." 
                                    - https://www.gamasutra.com/view/feature/3942/data_alignment_part_1.php?print=1

                              Notes from: https://www.youtube.com/watch?v=OKjOZBaKlOc
                                  - Data has to be aligned according to its length
                                  - Data that is 1 byte in sized can be saved anywhere in memory

                                  - addresses of 2 bytes of data (halfwords) must be multiples of 2 (even numbers)
                                        0, 2, 4, etc. - and not 1, 3, 5, etc.

                                  - addresses of 4 bytes of data (words) must be multiples of 4
                                  - addresses of 8 bytes of data (doublewords) must be multiples of 8


                                  Here is an example of aligning 4 bytes of data
                                  -----------
                                  0000 - it can start here
                                  0001 
                                  0002                      # The distance between each starting position is 4 addresses
                                  0003
                                  0004 - it can start here
                                  0005
                                  0006                      # Each one of these addresses is a byte
                                  0007
                                  0008 - it can start here
                                  0009
                                  000a 
                                  000b
                                  000c - it can start here
                                  000d
                                  000e
                                  000f
                                  0010 - it can start here
                                  0011 
                                  0012 
                                  0013 
                                  0014 - it can start here
                                  0015 
                                  0016 
                                  0017 
                                  0018 - it can start here
                                  0019 
                                  001a 
                                  001b 
                                  001c - it can start here
                                  001d 
                                  001e 
                                  001f            

        audio left off - These next ones control whether the audio is outputted in the left & right channels
        audio left on    The channels (based on my limted knowledge) only matter if you are using head/ear phones
        audio right off  These are not valid for MCIWAVE, MCISEQ, and Video Overlay deviced. No idea what this means,
        audio right on   but it doesn't work on my computer using the default speaker or headphones

        time format bytes - The time format of the audio, ie how do reference a point in time of the audio
        time format ms - the length of the audio in these units depends on the time format. 
        time format samples

    ----------------------------------------------------------------------------------------------------
    play: https://docs.microsoft.com/en-us/windows/win32/multimedia/play
    PLAY FLAGS FOR AUDIO:

        "play {alias} {flags & values}"

        from {position}
"""

"""
WAV File Format (http://soundfile.sapp.org/doc/WaveFormat/):

RIFF Chunk
    0 - ChunkID (4 bytes) - the letters 'RIFF'

    4 - ChunkSize (4 bytes) - "24 08 00 00" with the most significant byte being on the right. 
                              So converting 00 00 08 24 from hexadecimal to decimal gives 2084, which is the 
                              size of the rest of the RIFF following this byte. The total size of the chunk is then
                              8 bytes + this number which counts the ChunkID's & ChunkSize's 4 bytes each (8 bytes)

    8 - Format (4 bytes) - the letters 'WAVE'

fmt SubChunk

    12 - SubchunkID (4 bytes) - The letters 'fmt ' (note the space)

    Most significant byte on the right for the rest of the values in this subchunk

    16 - SubchunkSize (4 bytes) - 16 for PCM (Pulse-Code modulation)
    20 - AudioFormat (2 bytes) - 1 for PCM, otherwise there is some sort of compression

    22 - NumChannels (2 bytes) - Gives the number of channels in the audio. Mono is 1, Sterio is 2, etc.
    24 - SampleRate (4 bytes) - Number of samples per second
    28 - ByteRate (4 bytes) - SampleRate * NumChannels * BitsPerSample/8, the number of bytes to read per second

    32 - BlockAlign (2 bytes) - NumChannels * BitsPerSample/8, number of bytes for 1 sample including all channels

    34 - BitsPerSample (2 bytes)

data SubChunk

    36 - SubchunkID (4 bytes) - The letters 'data'

    Most significant byte on the right for the rest of the values in this subchunk

    40 - SubchunkSize (4 bytes) - The size of the data in the rest of this subchunk
    44 - Data (* bytes) - The actual sound data

Example WAV Audio File - eatghost.wav
----------------------------------------------------------------
RIFF Chunk

    RIFF\x04\x19\x00\x00WAVE

    0 ChunkID - RIFF
    4 ChunkSize - \x04\x19\x00\x00 = 00 00 19 04 (hex) = 6404 bytes
    8 Format - WAVE

fmt Subchunk

    I've added | between each field to help organize them. These are not part of the file.

    fmt \x10\x00\x00\x00|\x01\x00|\x01\x00|\x11\x2b\x00\x00|\x11\x2b\x00\x00|\x01\x00|\x08\x00

    12 SubchunkID - fmt 
    16 SubchunkSize - \x10\x00\x00\x00 = 00 00 00 10 (hex) = 16 bytes
    20 AudioFormat - \x01\x00 = 00 01 (hex) = 1 which stands for PCM
    22 NumChannels - \x01\x00 = 00 01 (hex) = 1 which means the audio is Mono
    24 SampleRate - \x11\x2b\x00\x00 = 00 00 2b 11 (hex) = 11025 Hz
    28 ByteRate - \x11\x2b\x00\x00 = 00 00 2b 11 (hex) = 11025 bytes/sec 
                                     which equals SampleRate * NumChannels * BitsPerSample/8
    32 Block Align - \x01\x00 = 00 01 (hex) = 1 byte/sample 
                                which equals NumChannels * BitsPerSample/8
    34 BitsPerSample - \x08\x00 = 00 08 (hex) = 8 bits/sample

data Subchunk   

    data\xb4\x18\x00\x00 

    36 SubchunkID - data
    40 SubchunkSize - \xb4\x18\x00\x00 = 00 00 18 b4 (hex) = 6324 bytes

    44 Data - The rest of the 6324 bytes are the data of the sound
"""

# The code for this function has pretty much (few changes) been copied from the playsound module
# https://github.com/TaylorSMarks/playsound
# Without the playsound module, I would have never known how to get started with playing sounds and it sort of paved
# the way for this part of goopylib.
def windows_command(command):
    buf = c_buffer(255)
    command = command.encode(getfilesystemencoding())

    error_code = windll.winmm.mciSendStringA(command, buf, 254, 0)

    if error_code:
        error_buffer = c_buffer(255)
        windll.winmm.mciGetErrorStringA(error_code, error_buffer, 254)
        raise GraphicsError(
            f'\n\tError {error_code} for command: \n\t\t{command.decode()}\n\t{error_buffer.value.decode()}')
    return buf.value


# SOME NOTES FOR GOOPY
# * The Alias names can be set by the user to reference certain sounds using a tag
# * There must be functions to get the different flags of the audio like number of channels & stuff
# * The user must be able to specify the time format type (for advanced purposes), default remains ms

# IMPORTANT: For Windows 2000/XP, send the stop command before closing

class Sound:
    file_instances = {*()}
    instances = {*()}

    def __init__(self, filepath, tag=None):
        tag = filepath if tag is None else tag
        i = 0
        self.tag = tag
        while self.tag in Sound.file_instances:
            self.tag = f"{tag}{i}"
            i += 1

        self.filepath = filepath

        windows_command(f'open "{self.filepath}" alias {self.tag}')
        windows_command(f'set {self.tag} time format samples')

        self.raw_duration = int(windows_command(f'status {self.tag} length').decode())

        self.is_playing = False
        self.is_paused = False

        self.start_crop = 0
        self.end_crop = 0

        Sound.file_instances.add(self.tag)
        Sound.instances.add(self)

    def __check_playing(self):
        if int(windows_command(f"status {self.tag} position").decode()) == \
                int(windows_command(f"status {self.tag} length").decode()):
            self.is_playing = False

    def play(self, wait_for_finish=False, start=0, end=0):
        windows_command(f"play {self.tag} from {start} to {self.raw_duration - end} notify")

        self.is_playing = True
        self.is_paused = False

        if wait_for_finish:
            sleep(self.raw_duration / 1000)

        return self

    def stop(self):
        windows_command(f"stop {self.tag}")

        self.is_playing = False
        self.is_paused = False

        return self

    def pause(self):
        self.__check_playing()
        if self.is_playing:
            windows_command(f"pause {self.tag}")
            self.is_paused = True
        return self

    def unpause(self):
        self.__check_playing()
        if self.is_playing:
            windows_command(f"resume {self.tag}")
            self.is_paused = False
        return self

    def crop(self, start=0, end=0):
        windows_command(f"delete {self.tag} from {self.raw_duration - end}")  # Cropping the sound
        windows_command(f"delete {self.tag} from 0 to {start}")

        self.raw_duration = int(windows_command(f'status {self.tag} length').decode())

        self.start_crop = start
        self.end_crop = end

        return self

    def save(self, filename):
        windows_command(f"save {self.tag} {filename}")
        return self

    def destroy(self):
        if self.tag in Sound.file_instances:
            windows_command(f"stop {self.tag}")
            windows_command(f"close {self.tag}")
            Sound.file_instances.remove(self.tag)

            self.is_playing = False
        return self

    # ------------------------------
    # GETTER FUNCTIONS

    def get_filepath(self):
        return self.tag

    def get_length(self):
        return int(windows_command(f"status {self.tag} length").decode())

    def get_bits_per_sample(self):
        return int(windows_command(f"status {self.tag} bitspersample").decode())

    def get_bytes_per_second(self):
        return int(windows_command(f"status {self.tag} bytespersec").decode())

    def get_number_of_channels(self):
        return int(windows_command(f"status {self.tag} channels").decode())

    def get_samples_per_second(self):
        return int(windows_command(f"status {self.tag} samplespersec").decode())

    def get_start_crop(self):
        return self.start_crop

    def get_end_crop(self):
        return self.end_crop

    def get_current_position(self):
        return int(windows_command(f"status {self.tag} position").decode())

    def get_percent_complete(self):
        return int(windows_command(f"status {self.tag} position").decode()) \
               / int(windows_command(f"status {self.tag} length").decode())

    def check_is_playing(self):
        if int(windows_command(f"status {self.tag} position").decode()) == \
                int(windows_command(f"status {self.tag} length").decode()):
            self.is_playing = False
        return self.is_playing and not self.is_paused



from goopylib.objects.Line import Line
from goopylib.math.Interpolations import *
from goopylib.constants import CURVE_INTERPOLATIONS

# The Curved Line class is used for drawing curved lines segments between 2 or more points
# The Curved line class is really just a line split into multiple segments and smoothed to give the illusion of a curve

class CurvedLine(Line):

    def __init__(self, *p, outline=None, outline_width=None, arrow=None, capstyle=None, joinstyle=None,
                 cursor="arrow", arrow_shape=None, arrow_scale=0.5, dash=None, interpolation="cosine", resolution=5,
                 smooth=True, bounds_width=None, layer=0, tag=None,
                 bias=0, tension=1):  # These last two are only required for hermite interpolation

        if not isinstance(resolution, int):
            raise GraphicsError(f"\n\nGraphicsError: The Resolution for the curved line must be an integer, "
                                f"not {resolution}")
        if resolution < 0:
            raise GraphicsError(f"\n\nGraphicsError: The Resolution for the curved line must be a non-negative integer, "
                                f"not {resolution}")
        if interpolation not in CURVE_INTERPOLATIONS:
            raise GraphicsError("\n\nGraphicsError: The Interpolation for the curved line must be one of "
                                f"{CURVE_INTERPOLATIONS}, not {interpolation}")

        if not (isinstance(bias, float) or isinstance(bias, int)):
            raise GraphicsError(f"\n\nGraphicsError: The bias for the curved line must be an integer or float, "
                                f"not {bias}")
        if not (isinstance(tension, float) or isinstance(tension, int)):
            raise GraphicsError(f"\n\nGraphicsError: The tension for the curved line must be an integer or float, "
                                f"not {tension}")

        self.is_smooth = smooth
        if interpolation == "linear":
            self.is_smooth = False

        self.points = list(p)
        self.points_copy = self.points.copy()

        if interpolation in ["cosine", "linear"]:
            for point in range(len(p) - 1):
                for t in range(1, resolution + 1):
                    tn = t / (resolution + 1)
                    x = self.points[point][0] + tn * (self.points[point + 1][0] - self.points[point][0])

                    if interpolation == "cosine":
                        new_point = [x, CosineInterpolation(self.points[point], self.points[point + 1], tn)]
                    else:
                        new_point = [x, LinearInterpolation(self.points[point], self.points[point + 1], tn)]

                    self.points_copy.insert((point * (resolution + 1)) + t, new_point)

        elif interpolation in ["hermite", "cubic"]:
            self.points.insert(0, self.points[0])
            self.points.insert(-1, self.points[-1])
            self.points_copy = self.points.copy()

            for point in range(1, len(self.points) - 2):
                for t in range(1, resolution + 1):
                    tn = t / (resolution + 1)
                    x = self.points[point][0] + tn * (self.points[point + 1][0] - self.points[point][0])

                    if interpolation == "cubic":
                        new_point = [x, CubicInterpolation(self.points[point - 1], self.points[point],
                                                           self.points[point + 1], self.points[point + 2], tn)]
                    else:
                        new_point = [x, HermiteInterpolation(self.points[point - 1], self.points[point],
                                                             self.points[point + 1], self.points[point + 2], tn,
                                                             tension, bias)]

                    self.points_copy.insert((point * (resolution + 1)) + t - resolution, new_point)

            self.points_copy = self.points_copy[1:-1]

        Line.__init__(self, *self.points_copy, outline=outline, outline_width=outline_width, arrow=arrow,
                      capstyle=capstyle, joinstyle=joinstyle, cursor=cursor, arrow_shape=arrow_shape, tag=tag,
                      arrow_scale=arrow_scale, dash=dash, bounds_width=bounds_width, layer=layer)

    def __repr__(self):
        return f"Curved Line({self.points})"

    def __dir__(self):
        return "see https://github.com/BhavyeMathur/goopylib/wiki/Lines-&-Curved-Lines!"

    def _draw(self, canvas, options):
        # Converting all the coordinates to Window coordinates to account for stretching, changed coords, etc.
        points = [canvas.to_screen(point[0], point[1]) for point in self.points]

        return canvas.create_line(points, options, smooth=self.is_smooth)  # Creating the line!



from goopylib.objects.GraphicsObject import GraphicsObject

from goopylib.styles import *


class BBox(GraphicsObject):
    # Internal base class for objects represented by bounding box
    # (opposite corners) Line segment is a degenerate case.

    resizing_objects = []

    def __init__(self, p1, p2, bounds=None, fill=None, outline=None, outline_width=None, cursor="arrow", layer=0,
                 tag=None):

        self.p1 = p1
        self.p2 = p2

        # These make sure that the p2 is 'after' p1, ie the x & y value of p2 is greater than that of p1
        if self.p1[0] > self.p2[0]:  # Checking if p1's x value is greater than p2's. If so, then swap the values
            self.p1[0], self.p2[0] = self.p2[0], self.p1[0]
        if self.p1[1] > self.p2[1]:  # Checking if p1's y value is greater than p2's. If so, then swap the values
            self.p1[1], self.p2[1] = self.p2[1], self.p1[1]

        self.anchor = [(self.p1[0] + self.p2[0]) // 2, (self.p1[1] + self.p2[1]) // 2]

        GraphicsObject.__init__(self, options=(), cursor=cursor, layer=layer, bounds=bounds, tag=tag)

        # abs(p2[0] - p1[0]) is not required because the p2 value is always greater than or equal to the p1 value
        self.width = self.p2[0] - self.p1[0]
        self.height = self.p2[1] - self.p1[1]

        self.min_width = None
        self.min_height = None
        self.max_width = None
        self.max_height = None

        self.resizing_bounds = {}
        self.is_resizing = {}
        self.bounds_thickness = 0

        if isinstance(fill, Colour):
            self.fill = fill
        else:
            self.fill = STYLES["default"]["fill"]

        if isinstance(outline, Colour):
            self.outline = outline
        else:
            self.outline = STYLES["default"]["outline"]

        if isinstance(outline_width, int):
            self.outline_width = outline_width
        else:
            self.outline_width = STYLES["default"]["width"]

    def __repr__(self):
        return "_BBox"

    def _set_resizable(self, resizables, top_bounds=None, bottom_bounds=None, left_bounds=None, right_bounds=None,
                       thickness=10):
        """Override in subclasses"""
        pass

    def _move(self, dx, dy):
        self.p1[0] += dx
        self.p1[1] += dy
        self.p2[0] += dx
        self.p2[1] += dy

        self.anchor[0] += dx
        self.anchor[1] += dy

    def get_p1(self):
        return self.p1.copy()

    def get_p2(self):
        return self.p2.copy()

    def is_clicked(self, mouse_pos):
        if self.bounds is None:
            if mouse_pos is None:
                return False
            else:
                if (self.p1[0] < mouse_pos[0] < self.p2[0] or self.p1[0] > mouse_pos[0] > self.p2[0]) and \
                        (self.p1[1] < mouse_pos[1] < self.p2[1] or self.p1[1] > mouse_pos[1] > self.p2[1]):
                    return True
                else:
                    return False
        else:
            return self.bounds.is_clicked(mouse_pos)

    def set_coords(self, p1, p2):

        self.p1 = p1.copy()
        self.p2 = p2.copy()

        # These make sure that the p2 is 'after' p1, ie the x & y value of p2 is greater than that of p1
        if self.p1[0] > self.p2[0]:  # Checking if p1's x value is greater than p2's. If so, then swap the values
            self.p1[0], self.p2[0] = self.p2[0], self.p1[0]
        if self.p1[1] > self.p2[1]:  # Checking if p1's y value is greater than p2's. If so, then swap the values
            self.p1[1], self.p2[1] = self.p2[1], self.p1[1]

        # abs(p2[0] - p1[0]) is not required because the p2 value is always greater than or equal to the p1 value
        self.width = self.p2[0] - self.p1[0]
        self.height = self.p2[1] - self.p1[1]
        width_scale = (p2[0] - p1[0]) / self.width
        height_scale = (p2[1] - p1[1]) / self.height

        # abs(p2[0] - p1[0]) is not required because the p2 value is always greater than or equal to the p1 value
        self.width = p2[0] - p1[0]
        self.height = p2[1] - p1[1]

        self.anchor = [(self.p1[0] + self.p2[0]) // 2, (self.p1[1] + self.p2[1]) // 2]

        self._update_layer()

        return self

    def set_width(self, width, center="center"):
        if center not in {"center", "right", "left"}:
            raise GraphicsError(
                "\n\nThe center argument for resizing the object (set_outline_width) needs to be one of "
                f'{["center", "right", "left"]}')

        if center == "left":
            self.set_coords(self.p1, self.p2.add_x(width - self.width))

        elif center == "right":
            self.set_coords(self.p1.add_x(-(width - self.width)), self.p2)
        else:
            self.set_coords(self.p1.add_x(-(width / 2 - self.width)), self.p2.add_x(width / 2 - self.width))

        return self

    def set_height(self, height, center="center"):
        if center not in {"center", "top", "bottom"}:
            raise GraphicsError(
                "\n\nThe center argument for resizing the object (set_height) needs to be one of "
                f'{["center", "top", "bottom"]}')

        if center == "top":
            self.set_coords(self.p1, self.p2.add_y(height - self.height))

        elif center == "bottom":
            self.set_coords(self.p1.add_y(-(height - self.height)), self.p2)
        else:
            self.set_coords(self.p1.add_y(-(height / 2 - self.height)), self.p2.add_y(height / 2 - self.height))

        return self

    def get_width(self):
        return self.width

    def get_height(self):
        return self.height

    def get_fill(self):
        return self.fill

    def get_outline(self):
        return self.outline

    def get_outline_width(self):
        return self.outline_width

    def set_dimensions(self, width, height, horizontal_align="center", vertical_align="center"):
        self.set_width(width, horizontal_align)
        self.set_height(height, vertical_align)
        return self

    def set_resizable(self, top=False, left=False, bottom=False, right=False, min_width=40, min_height=40,
                      bounds_width=10, top_bounds=None, bottom_bounds=None, left_bounds=None, right_bounds=None):

        if min_width < 1 or min_height < 1:
            raise GraphicsError(f"\n\nGraphicsError: Minimum height and width of resizable object must be greater than "
                                f"or equal to 1. Right now, min_width={min_width} & min_height={min_height}")

        self.min_width = min_width
        self.min_height = min_height

        self.is_resizing = {"top": top, "left": left, "bottom": bottom, "right": right}

        self._set_resizable([top, bottom, left, right], top_bounds=top_bounds, bottom_bounds=bottom_bounds,
                            left_bounds=left_bounds, right_bounds=right_bounds, thickness=bounds_width)

        if top is False and bottom is False and left is False and right is False:
            if self in GraphicsObject.resizing_objects:
                GraphicsObject.resizing_objects.remove(self)

        elif self not in GraphicsObject.resizing_objects:
            GraphicsObject.resizing_objects.add(self)

        self.bounds_thickness = bounds_width

        return self

    def get_anchor(self):
        return self.anchor



from goopylib.math.PyBezierCurve import *

try:
    from goopylib.math.CBezierCurve import *
except ImportError as e:
    print(f"Failed to import goopylib.math.CBezierCurve, Post this error on github.com/BhavyeMathur/goopylib/issues: {e}.")

from goopylib.math.BSpline import *

from goopylib.math.PyEasing import *

from goopylib.math.Triangulation import *

from goopylib.math.Curves import *
from goopylib.math.Interpolations import *



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.styles import *

from goopylib.math.Triangulation import triangulate_modified_earclip
from math import radians, sin, cos

class Polygon(GraphicsObject):
    def __init__(self, *points, style=None, fill=None, outline=None, outline_width=None, is_rounded=False, roundness=5,
                 layer=0, tag=None):

        # if points passed as a list, extract it
        if len(points) == 1 and isinstance(points[0], list):
            points = points[0]
        self.points = points
        if len(self.points) < 3:
            raise GraphicsError(f"\n\nGraphicsError: there must be at least 3 points for the polygon")

        self.is_rounded = is_rounded
        self.sharpness = roundness

        self.low_x, self.low_y = None, None  # These variables are all defined in the get_size() function called below
        self.high_x, self.high_y = None, None  # They store the extreme x & y points farthest from the line's anchor
        self.width, self.height = None, None

        self.calculate_dimensions()

        GraphicsObject.__init__(self, options=("fill", "outline", "width"), layer=layer, tag=tag)

        if isinstance(fill, Colour):
            self.fill = fill
        elif fill in STYLES[self.style].keys():
            self.fill = STYLES[self.style][fill]
        else:
            if "fill" in STYLES[self.style].keys():
                self.fill = STYLES[self.style]["fill"]
            else:
                self.fill = STYLES["default"]["fill"]
        self.set_fill(self.fill)

        if isinstance(outline, Colour):
            self.outline = outline
        elif outline in STYLES[self.style].keys():
            self.outline = STYLES[self.style][outline]
        else:
            if "outline" in STYLES[self.style].keys():
                self.outline = STYLES[self.style]["outline"]
            else:
                self.outline = STYLES["default"]["outlien"]
        self.set_outline(self.outline)

        if isinstance(outline_width, int):
            self.outline_width = outline_width
        elif outline_width in STYLES[self.style].keys():
            self.outline_width = STYLES[self.style][outline_width]
        else:
            if "width" in STYLES[self.style].keys():
                self.outline_width = STYLES[self.style]["width"]
            else:
                self.outline_width = STYLES["default"]["width"]
        self.set_outline_width(self.outline_width)

        self.triangles = None
        self.triangulate()

    def __repr__(self):
        return f"Polygon({self.points})"

    def _draw(self, canvas, options):
        points = []
        for p in self.points:
            x, y = canvas.to_screen(p.x, p.y)
            points.append(x)
            points.append(y)

        if options["width"] == 0:
            options["outline"] = options["fill"]
        return canvas.create_polygon(points, options, smooth=self.is_rounded)

    def _move(self, dx, dy):
        for p in self.points:
            p.move(dx, dy)

    def _rotate(self, dr, sampling="bicubic", center=None):
        if center is None:
            center = self.get_anchor()

        self.move(-center.x, -center.y)

        points = []
        for p in self.points:
            points.append(Point(cos(radians(dr)) * p.x - sin(radians(dr)) * p.y + center.x,
                                sin(radians(dr)) * p.x + cos(radians(dr)) * p.y + center.y))
        self.points = points

    def clone(self):
        other = Polygon(*self.points)
        other.config = self.config.copy()
        return other

    def flip_x(self):
        return self

    def get_points(self):
        return list(map(Point.clone, self.points))

    def calculate_dimensions(self):
        # To calculate the width, it finds the smallest & largest x points and gets the width
        self.low_x = self.points[0].x  # Setting the High & Low point to the first point
        self.high_x = self.low_x

        for point in self.points[1:]:  # Going through all the other points
            if point.x < self.low_x:  # Checking if this point is smaller
                self.low_x = point.x
            elif point.x > self.high_x:  # Checking if this point is larger
                self.high_x = point.x
        self.width = abs(self.high_x - self.low_x)  # Getting the width

        # To calculate the height, it finds the smallest & largest y points and gets the height
        self.low_y = self.points[0].y  # Setting the High & Low point to the first point
        self.high_y = self.low_y

        for point in self.points[1:]:  # Going through all the other points
            if point.y < self.low_y:  # Checking if this point is smaller
                self.low_y = point.y
            elif point.y > self.high_y:  # Checking if this point is larger
                self.high_y = point.y
        self.height = abs(self.high_y - self.low_y)  # Getting the height

    def get_width(self):
        return self.width

    def get_height(self):
        return self.height

    def is_clicked(self, pos):
        if pos is None:
            return False
        else:
            if self.bounds is None:
                width = self.get_width() / 2
                height = self.get_height() / 2
    
                # Check if point in Rectangular Bounding Box
                if self.anchor.x - width < pos.x < self.anchor.x + width and \
                   self.anchor.y - height < pos.y < self.anchor.y + height:
    
                    """
                    # Ray Casting Algorithm
                    # Code Modified from
                    # https://stackoverflow.com/questions/217578/how-can-i-determine-whether-a-2d-point-is-within-a-polygon
                    intersections = 0
                    size = len(self.points)
    
                    v2p1 = (self.anchor.x - width - 1, pos.y)
                    for side in range(size):
                        v1p1 = self.points[side]
                        v1p2 = self.points[(side + 1) % size]
    
                        a1 = v1p2.y - v1p1.y
                        b1 = v1p1.x - v1p2.x
                        c1 = (v1p2.x * v1p1.y) - (v1p1.x * v1p2.y)
    
                        d1 = (a1 * pos.x) + (b1 * pos.y) + c1
                        d2 = (a1 * v2p1[0]) + (b1 * v2p1[1]) + c1
                        if d1 == 0 and d2 == 0:
                            a2 = v2p1[1] - pos.y
                            b2 = pos.x - v2p1[0]
                            c2 = (v2p1[0] * pos.y) - (pos.x * v2p1[1])
    
                            d1 = (a2 * v1p1.x) + (b2 * v1p1.y) + c2
                            d2 = (a2 * v1p2.x) + (b2 * v1p2.y) + c2
    
                            if d1 == 0 and d2 == 0:
                                intersections += 1
    
                    return intersections % 1 == 1
                    """
                    if len(self.points) == 3:
                        p1, p2, p3 = self.points
                        d1 = (pos.x - p2.x) * (p1.y - p2.y) - (p1.x - p2.x) * (pos.y - p2.y)
                        d2 = (pos.x - p3.x) * (p2.y - p3.y) - (p2.x - p3.x) * (pos.y - p3.y)
                        d3 = (pos.x - p1.x) * (p3.y - p1.y) - (p3.x - p1.x) * (pos.y - p1.y)

                        return not ((d1 < 0 or d2 < 0 or d3 < 0) and (d1 > 0 or d2 > 0 or d3 > 0))
                    else:
                        for triangle in self.triangles:
                            if triangle.is_clicked(pos):
                                return True
                else:
                    return False
            else:
                return self.bounds.is_clicked(pos)

    def get_anchor(self):
        x = 0
        y = 0
        for point in self.points:
            x += point.x
            y += point.y

        x /= len(self.points)
        y /= len(self.points)

        return Point(x, y)

    def triangulate(self):
        if self.triangles is None and len(self.points) != 3:
            self.triangles = []
            for triangle in triangulate_modified_earclip(list(self.points)):
                self.triangles.append(Polygon(*triangle))



from goopylib.Point import Point
from goopylib.util import GraphicsError
import math

def LinearCurve(t, control_points):
    if not len(control_points) > 2:
        raise GraphicsError("Length of Control Points must be greater than 2")

    total_distance = control_points[-1].x - control_points[0].x
    time_x = t * total_distance + control_points[0].x

    for i, point in enumerate(control_points):
        if time_x < point.x:
            lower_point = i - 1
            break

    upper_point = lower_point + 1
    points_distance = control_points[lower_point].distance_x(control_points[upper_point])

    x = (1 - t) * control_points[lower_point].x + t * control_points[upper_point].x
    proportion_done = (time_x - control_points[lower_point].x) / points_distance

    return Point(x, (1 - proportion_done) * control_points[lower_point].y + proportion_done * control_points[upper_point].y)

def CosineCurve(t, control_points):
    if not len(control_points) > 2:
        raise GraphicsError("Length of Control Points must be greater than 2")

    total_distance = control_points[-1].x - control_points[0].x
    time_x = t * total_distance + control_points[0].x

    for i, point in enumerate(control_points):
        if time_x < point.x:
            lower_point = i - 1
            break

    upper_point = lower_point + 1
    points_distance = control_points[lower_point].distance_x(control_points[upper_point])

    x = (1 - t) * control_points[lower_point].x + t * control_points[upper_point].x
    proportion_done = (time_x - control_points[lower_point].x) / points_distance

    ft = t * math.pi
    f = (1 - math.cos(ft)) / 2

    return Point(x, lower_point*(1-f) + upper_point*f)



import goopylib.math.CBezierCurve
from goopylib.util import GraphicsError

def combination(n, k):
    if isinstance(n, int):
        if isinstance(k, int):
            if n + 1 > k:  # faster than n >= k
                return goopylib.math.CBezierCurve.combination(n, k)
            else:
                raise GraphicsError(f"\n\nGraphicsError: n must be >= k, not n={n} & k={k}")
        else:
            raise GraphicsError(f"\n\nGraphicsError: k argument for combination must be an integer, not {k}")
    else:
        raise GraphicsError(f"\n\nGraphicsError: n argument for combination must be an integer, not {n}")

def bernstein_polynomial(i, n, t):
    if isinstance(t, (float, int)):
        if isinstance(i, int):
            if isinstance(n, int):
                if n + 1 > i:  # Faster than n >= 1
                    return goopylib.math.CBezierCurve.bernstein_polynomial(i, n, t)
                else:
                    raise GraphicsError(f"\n\nGraphicsError: n must be >= i, not n={n} & i={i}")
            else:
                raise GraphicsError(f"\n\nGraphicsError: n argument for bernstein polynomial must be an int, not {n}")
        else:
            raise GraphicsError(f"\n\nGraphicsError: i argument for bernstein polynomial must be an integer, not {i}")
    else:
        raise GraphicsError(f"\n\nGraphicsError: t parameter for bernstein polynomial must be a float or int, not {t}")

def bezier_curve(t, control_points):
    if isinstance(t, (float, int)):
        return goopylib.math.CBezierCurve.bezier_curve(t, control_points)
    else:
        raise GraphicsError(f"\n\nGraphicsError: t parameter for bezier curve must be a float or int, not {t}")

def rational_bezier_curve(t, control_points, weights):
    if isinstance(t, (float, int)):
        return goopylib.math.CBezierCurve.rational_bezier_curve(t, control_points, weights)
    else:
        raise GraphicsError("\n\nGraphicsError: t parameter for rational bezier curve must be a float or int, "
                            f"not {t}")



from goopylib.objects.Checkbox import Checkbox
from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.util import GraphicsError

class RadioButton(GraphicsObject):

    def __init__(self, *checkboxes, state=0, cursor="arrow", layer=0, tag=None, autoflush=True):
        if len(checkboxes) > 0:
            for obj in checkboxes:
                if not isinstance(obj, Checkbox):
                    raise GraphicsError(f"\n\nGraphicsError: All Radio Button states must be Checkboxes, not {obj}")
                if obj in GraphicsObject.cyclebutton_instances:
                    GraphicsObject.cyclebutton_instances.remove(obj)
                obj.set_state(False)

            self.checkboxes = checkboxes

            super().__init__(options=(), cursor=cursor, layer=layer, tag=tag)
            if autoflush:
                GraphicsObject.radiobutton_instances.add(self)

            self.state = state

            number_of_states = 0
            self.anchor = [0, 0]
            for state in self.checkboxes:
                state_anchor = state.get_anchor()
                if state_anchor is not None:
                    self.anchor[0] += state_anchor[0]
                    self.anchor[1] += state_anchor[1]
                    number_of_states += 1

            if number_of_states != 0:
                self.anchor = [self.anchor[0] // number_of_states, self.anchor[1] // number_of_states]

            self.current_state = self.checkboxes[self.state].set_state(True)
        else:
            raise GraphicsError("\n\nGraphicsError: RadioButton must have at least 1 state, not 0")

    def __repr__(self):
        return f"RadioButton({self.checkboxes[self.state]} and {len(self.checkboxes) - 1} others)"

    def __iter__(self):
        for state in self.checkboxes:
            yield state

    def __len__(self):
        return len(self.checkboxes)

    def __getitem__(self, item):
        return self.checkboxes[item]

    def _draw(self, canvas, options):
        for checkbox in self.checkboxes:
            checkbox.draw(canvas, _internal_call=True)
        return self

    def _undraw(self):
        for checkbox in self.checkboxes:
            checkbox.undraw()

    def destroy(self):
        GraphicsObject.objects.discard(self)
        GraphicsObject.object_layers[self.layer].discard(self)
        GraphicsObject.draggable_objects.discard(self)
        GraphicsObject.cursor_objects.discard(self)

        GraphicsObject.radiobutton_instances.discard(self)

        if self in GraphicsObject.redraw_on_frame[self.layer]:
            GraphicsObject.redraw_on_frame[self.layer].remove(self)

        for checkbox in self.checkboxes:
            checkbox.destroy()

        self.drawn = False
        self.graphwin = None
        self.id = None

        return self

    def base_undraw(self):
        for checkbox in self.checkboxes:
            checkbox.base_undraw()

    def _rotate(self, dr, sampling="bicubic", center=None):
        for checkbox in self.checkboxes:
            checkbox.rotate(dr, sampling=sampling, center=center)

    def _move(self, dx, dy):
        self.anchor[0] += dx
        self.anchor[1] += dy
        for checkbox in self.checkboxes:
            checkbox.move(dx, dy)

    def get_anchor(self):
        return self.anchor

    def is_clicked(self, mouse_pos):
        if self.bounds is None:
            for checkbox in self.checkboxes:
                if checkbox.is_clicked(mouse_pos):
                    return True
        else:
            return self.bounds.is_clicked(mouse_pos)

        return False

    def set_autoflush(self, autoflush):
        if autoflush:
            if self not in GraphicsObject.radiobutton_instances:
                GraphicsObject.radiobutton_instances.add(self)
        elif self in GraphicsObject.radiobutton_instances:
            GraphicsObject.radiobutton_instances.remove(self)
        return self

    def set_state(self, state):
        for checkbox in self.checkboxes:
            if self.checkboxes[state] == checkbox:
                checkbox.click()
                self.current_state.click()
                self.current_state = checkbox
                break

    def get_state(self):
        return self.checkboxes.index(self.current_state)

    def get_object(self):
        return self.current_state



from goopylib.Point import Point
from goopylib.objects.GraphicsObject import GraphicsObject

import math

class SolidArc(GraphicsObject):
    def __init__(self, p, start_angle, end_angle, radius, radius2=None, outline=None, outline_width=None,
                 cursor="arrow", arrow=None, resolution=10, smooth=True, layer=0, tag=None):

        self.start_angle = start_angle
        self.end_angle = end_angle

        self.anchor = p

        self.radius1 = radius
        if radius2 is None:
            self.radius2 = radius
        else:
            self.radius2 = radius2

        angle_change = (end_angle - start_angle) / resolution

        range_end = 90 / angle_change
        decimal_points = min([len(str(range_end).split('.')[1]), 3])

        x_coeff = sum([abs(math.cos(math.radians(i * angle_change + start_angle)))
                       for i in range(int(range_end * 10 ** decimal_points))])
        y_coeff = sum([abs(math.sin(math.radians(i * angle_change + start_angle)))
                       for i in range(int(range_end * 10 ** decimal_points))])

        x_coeff /= 10 ** decimal_points
        y_coeff /= 10 ** decimal_points

        angle_change = (end_angle - start_angle) / resolution

        x_change = self.radius1 / 2 / x_coeff
        y_change = self.radius2 / 2 / y_coeff

        self.points = [p - Point(x_change/2, self.radius2 / 2)]

        for i in range(resolution):
            cur_angle = (i * angle_change) + start_angle
            self.points.append(self.points[-1] + Point(x_change * math.cos(math.radians(cur_angle)),
                                                       y_change * math.sin(math.radians(cur_angle))))

        GraphicsObject.__init__(self, options=["outline", "width", "fill"], tag=tag)

    def _draw(self, canvas, options):
        return canvas.create_polygon(self.points, options)



import time
import os
import sys

from tkinter.filedialog import askopenfilename

from goopylib.constants import _root, _update_lasttime

class GraphicsError(Exception):
    """Generic error class for graphics module exceptions."""
    pass


class GraphicsWarning(Warning):
    pass

# Other Image Functions

def resource_path(relative_path):
    # Get absolute path to resource, works for dev and for PyInstaller
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except AttributeError:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)


# Time Related Functions

def update(rate=None):
    global _update_lasttime
    if rate:
        now = time.time()
        pauseLength = 1 / rate - (now - _update_lasttime)
        if pauseLength > 0:
            time.sleep(pauseLength)
            _update_lasttime = now + pauseLength
        else:
            _update_lasttime = now

    _root.update()


def get_screen_size():
    return _root.winfo_screenwidth(), _root.winfo_screenheight()

def get_screen_height():
    return _root.winfo_screenheight()

def get_screen_width():
    return _root.winfo_screenwidth()

def openfilebrowser():
    return askopenfilename()



from goopylib.util import GraphicsError

class Point:
    def __init__(self, x, y):

        if not ((isinstance(x, int) or isinstance(x, float)) and (isinstance(y, int) or isinstance(y, float))):
            raise GraphicsError("\n\nx (x={}) & y (y={}) positions must be integers".format(x, y))

        self.x = x
        self.y = y

    def __repr__(self):
        return f"Point({self.x}, {self.y})"

    def __mul__(self, other):
        if isinstance(other, Point):
            return Point(self.x * other.x, self.y * other.y)
        elif isinstance(other, int) or isinstance(other, float):
            return Point(self.x * other, self.y * other)
        else:
            raise TypeError(f"unsupported operand type(s) for *: 'Point' and {type(other)}'")

    def __add__(self, other):
        if isinstance(other, Point):
            return Point(self.x + other.x, self.y + other.y)
        elif isinstance(other, int) or isinstance(other, float):
            return Point(self.x + other, self.y + other)
        else:
            raise TypeError(f"unsupported operand type(s) for +: 'Point' and {type(other)}'")

    def __sub__(self, other):
        if isinstance(other, Point):
            return Point(self.x - other.x, self.y - other.y)
        elif isinstance(other, int) or isinstance(other, float):
            return Point(self.x - other, self.y - other)
        else:
            raise TypeError(f"unsupported operand type(s) for -: 'Point' and {type(other)}'")

    def __truediv__(self, other):
        if isinstance(other, Point):
            return Point(self.x / other.x, self.y / other.y)
        elif isinstance(other, int) or isinstance(other, float):
            return Point(self.x / other, self.y / other)
        else:
            raise TypeError(f"unsupported operand type(s) for /: 'Point' and {type(other)}'")
        
    def __floordiv__(self, other):
        if isinstance(other, Point):
            return Point(self.x // other.x, self.y // other.y)
        elif isinstance(other, int) or isinstance(other, float):
            return Point(self.x // other, self.y // other)
        else:
            raise TypeError(f"unsupported operand type(s) for //: 'Point' and {type(other)}'")
        
    def __pow__(self, power, modulo=None):
        if isinstance(power, Point):
            return Point(pow(self.x, power.x, modulo), pow(self.y, power.y, modulo))
        elif isinstance(power, int) or isinstance(power, float):
            return Point(pow(self.x, power, modulo), pow(self.y, power, modulo))
        else:
            raise TypeError(f"unsupported operand type(s) for **: 'Point' and {type(power)}'")

    def __lt__(self, other):
        if isinstance(other, Point):
            return abs(self.distance(Point(0, 0))) < abs(other.distance(Point(0, 0)))
        else:
            raise TypeError(f"'<' not supported between instances of 'Point' and '{type(other)}'")

    def __pos__(self):
        return Point(abs(self.x), abs(self.y))

    def __lshift__(self, other):
        if isinstance(other, Point):
            return Point(int(self.x) << int(other.x), int(self.y) << int(other.y))
        elif isinstance(other, int) or isinstance(other, float):
            return self.clone() if other % 2 == 0 else Point(self.y, self.x)
        else:
            raise TypeError(f"unsupported operand type(s) for <<: 'Point' and {type(other)}'")

    def __rshift__(self, other):
        if isinstance(other, Point):
            return Point(int(self.x) >> int(other.x), int(self.y) >> int(other.y))
        elif isinstance(other, int) or isinstance(other, float):
            return self << other
        else:
            raise TypeError(f"unsupported operand type(s) for >>: 'Point' and {type(other)}'")

    def __copy__(self):
        return Point(self.x, self.y)

    def __bytes__(self):
        return bytes(int(self.x)), bytes(int(self.y))

    def __bool__(self):
        return self.x != 0 and self.y != 0

    def __and__(self, other):
        if isinstance(other, Point):
            return Point(int(self.x) & int(other.x), int(self.y) & int(other.y))
        elif isinstance(other, int) or isinstance(other, float):
            return Point(int(self.x) & int(other), int(self.y) & int(other))
        else:
            raise TypeError(f"unsupported operand type(s) for &: 'Point' and {type(other)}'")

    def __abs__(self):
        return Point(abs(self.x), abs(self.y))

    def __ceil__(self):
        return -(-self//1)

    def __floor__(self):
        return self//1

    def __neg__(self):
        return Point(-self.x, -self.y)

    def __mod__(self, other):
        if isinstance(other, Point):
            return Point(self.x % other.x, self.y % other.y)
        elif isinstance(other, int) or isinstance(other, float):
            return Point(self.x % other, self.y % other)
        else:
            raise TypeError(f"unsupported operand type(s) for %: 'Point' and {type(other)}'")

    def __divmod__(self, other):
        return self // other, self % other

    def __eq__(self, other):
        if isinstance(other, Point):
            return self.x == other.x and self.y == other.y
        else:
            return self.x == self.y == other

    def __ne__(self, other):
        return not self == other

    def __ge__(self, other):
        return self > other or self == other

    def __gt__(self, other):
        return not (self < other or self == other)

    def __iadd__(self, other):
        self.x, self.y = self + other
        return self

    def __hex__(self):
        return hex(int(self.x)), hex(int(self.y))

    def __ilshift__(self, other):
        self.x, self.y = self << other
        return self

    def __irshift__(self, other):
        self.x, self.y = self >> other
        return self

    def __imod__(self, other):
        self.x, self.y = self % other
        return self

    def __imul__(self, other):
        self.x, self.y = self * other
        return self

    def __invert__(self):
        return Point(self.y, self.x)

    def __or__(self, other):
        if isinstance(other, Point):
            return Point(int(self.x) | int(other.x), int(self.y) | int(other.y))
        elif isinstance(other, int) or isinstance(other, float):
            return Point(int(self.x) | int(other), int(self.y) | int(other))
        else:
            raise TypeError(f"unsupported operand type(s) for |: 'Point' and {type(other)}'")

    def __ior__(self, other):
        self.x, self.y = self | other
        return self

    def __ipow__(self, other):
        self.x, self.y = self ** other
        return self

    def __isub__(self, other):
        self.x, self.y = self - other
        return self

    def __iand__(self, other):
        self.x, self.y = self & other
        return self

    def __ifloordiv__(self, other):
        self.x, self.y = self // other
        return self

    def __dir__(self):
        return "see https://github.com/BhavyeMathur/goopylib/wiki"

    def __contains__(self, item):
        return item == self.x or item == self.y

    def __ixor__(self, other):
        self.x, self.y = self ^ other
        return self
        
    def __xor__(self, other):
        if isinstance(other, Point):
            return Point(int(self.x) ^ int(other.x), int(self.y) ^ int(other.y))
        elif isinstance(other, int) or isinstance(other, float):
            return Point(int(self.x) ^ int(other), int(self.y) ^ int(other))
        else:
            raise TypeError(f"unsupported operand type(s) for ^: 'Point' and {type(other)}'")

    def __round__(self, n=None):
        return Point(round(self.x, n), round(self.y, n))

    def __len__(self):
        return 2

    def __getitem__(self, item):
        return (self.x, self.y)[item]

    def __iter__(self):
        for x in (self.x, self.y):
            yield x

    def __le__(self, other):
        return self < other or self == other

    def __itruediv__(self, other):
        self.x, self.y = self / other
        return self

    def __oct__(self):
        return oct(int(self.x)), oct(int(self.y))

    def __reversed__(self):
        return Point(self.y, self.x)

    def __deepcopy__(self, memodict={}):
        return Point(self.x, self.y)

    def distance(self, p2):
        if not isinstance(p2, Point):
            raise GraphicsError(f"\n\np2 argument for distance calculation must be a Point class, not {p2}")
        return ((p2.x - self.x) ** 2 + (p2.y - self.y) ** 2) ** 0.5

    def distance_x(self, p2):
        if not isinstance(p2, Point):
            raise GraphicsError(f"\n\np2 argument for distance calculation must be a Point class, not {p2}")
        return p2.x - self.x

    def distance_y(self, p2):
        if not isinstance(p2, Point):
            raise GraphicsError(f"\n\np2 argument for distance calculation must be a Point class, not {p2}")
        return p2.y - self.y

    def slope(self, p2):
        if not isinstance(p2, Point):
            raise GraphicsError(f"\n\np2 argument for slope calculation must be a Point class, not {p2}")
        if self.x == p2.x:
            raise GraphicsError(f"\n\nThe two points have the same x value, cannot calculate slope!")
        return (self.y - p2.y) / (self.x - p2.x)

    def clone(self):
        return Point(self.x, self.y)

    def swap(self):
        self.x, self.y = self.y, self.x

    def get_x(self):
        return self.x

    def get_y(self):
        return self.y

    def move_y(self, dy):
        if not (isinstance(dy, int) or isinstance(dy, float)):
            raise GraphicsError("\n\nThe value to move the point by (dy) must be a number (integer or float), "
                                f"not {dy}")
        self.y += dy
        return self

    def move_x(self, dx):
        if not (isinstance(dx, int) or isinstance(dx, float)):
            raise GraphicsError("\n\nThe value to move the point by (dx) must be a number (integer or float), "
                                f"not {dx}")
        self.x += dx
        return self

    def move(self, dx, dy):
        self.move_y(dy)
        self.move_x(dx)
        return self

    def move_to(self, x, y):
        self.move_to_x(x)
        self.move_to_y(y)
        return self

    def move_to_x(self, x):
        if not (isinstance(x, int) or isinstance(x, float)):
            raise GraphicsError(f"\n\nThe value to move the point by (x) must be a number (integer or float), not {x}")
        self.x = x
        return self

    def move_to_y(self, y):
        if not (isinstance(y, int) or isinstance(y, float)):
            raise GraphicsError(f"\n\nThe value to move the point by (y) must be a number (integer or float), not {y}")
        self.y = y
        return self

    def add(self, dx, dy):
        return self.clone().move(dx, dy)

    def set(self, x, y):
        return self.clone().move_to(x, y)

    def get_anchor(self):
        return self

    def add_y(self, dy):
        return self.clone().move_y(dy)

    def add_x(self, dx):
        return self.clone().move_x(dx)

    def set_x(self, x):
        return self.clone().move_to_x(x)

    def set_y(self, y):
        return self.clone().move_to_y(y)



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.styles import *
from goopylib.constants import _root, ALL_CHARACTERS, ALIGN_OPTIONS

from tkinter import StringVar as tkStringVar
from tkinter import Entry as tkEntry
from tkinter import Frame as tkFrame
from tkinter import END as tkEND

class Entry(GraphicsObject):

    def __init__(self, p, text_width=10, fill=None, font_colour=None, font_face=None, font_size=None,
                 outline=None, font_style=None, outline_width=None, border_relief="flat", password=False, layer=0,
                 active="NORMAL", justify="left", cursor="xterm", select_colour=None, prompt_text="", tag=None,
                 bounds=None, align="center", text="", max_characters=None):

        self.anchor = p.copy()

        if not isinstance(text_width, int):
            raise GraphicsError(f"Text Width for the Entry must be an integer, not {text_width}")
        self.text_width = text_width

        if align not in ALIGN_OPTIONS:
            raise GraphicsError(f"\n\nGraphicsError: Image align must be one of {ALIGN_OPTIONS}, not {align}")
        self.align = align

        self.text = tkStringVar(_root)
        self.text.set(text)

        self.was_edited_since_last_call = False

        if isinstance(fill, Colour):
            self.fill = fill
        else:
            self.fill = STYLES["default"]["fill"]

        if isinstance(select_colour, Colour):
            self.select_colour = select_colour
        else:
            self.select_colour = STYLES["default"]["select colour"]

        if isinstance(font_colour, Colour):
            self.font_colour = font_colour
        else:
            self.font_colour = STYLES["default"]["font colour"]

        if isinstance(outline_width, int):
            self.outline_width = outline_width
        else:
            self.outline_width = STYLES["default"]["entry width"]

        if isinstance(font_size, int):
            self.font_size = font_size
        else:
            self.font_size = STYLES["default"]["font size"]

        if isinstance(font_style, str):
            self.font_style = font_style
        else:
            self.font_style = STYLES["default"]["font style"]

        if isinstance(font_face, str):
            self.font = font_face
        else:
            self.font = STYLES["default"]["font face"]

        if isinstance(justify, str):
            self.justify = justify
        else:
            self.justify = STYLES["default"]["justify"]

        self.entry = None
        self.border_type = border_relief
        if password:
            self.text_type = "*"
        else:
            self.text_type = ""
        self.enabled = "normal"
        self.cursor = cursor
        self.prompt_text = prompt_text

        self.initial_font_size = self.font_size
        self.edited = False

        self.allowed_symbols = ALL_CHARACTERS
        self.max_characters = None
        self.set_max_characters(max_characters)

        self.last_graphwin = None

        GraphicsObject.__init__(self, (), layer=layer, tag=tag, bounds=bounds)

    def __repr__(self):
        if self.drawn:
            return f"Entry({self.anchor}, {self.get_text()})"
        else:
            return f"Entry({self.anchor})"

    def _draw(self, canvas, options):
        x, y = self.anchor

        if self.last_graphwin != canvas:
            self.frm = tkFrame(canvas.master)
            self.set_font_size(int(self.initial_font_size / canvas.trans.x_scale), False)

            self.last_graphwin = canvas

            self.entry = tkEntry(self.frm, width=self.text_width, textvariable=self.text, bg=self.fill, fg=self.font_colour,
                                 bd=self.outline_width, font=(self.font, self.font_size, self.font_style),
                                 insertbackground=self.font_colour, show=self.text_type, state=self.enabled,
                                 justify=self.justify, cursor="xterm", exportselection=0,
                                 selectbackground=self.select_colour, insertborderwidth=0)

        if not self.edited:
            self.entry.insert(0, self.prompt_text)
            self.entry.config(show="")
        self.entry.pack()

        self.entry.bind("<Return>", self._on_enter)

        self.text.trace('w', self._on_edit)

        width = self.get_width()
        height = self.get_height()

        if self.align == "center":
            pass
        elif self.align == "bottom":
            y -= self.font_size / 2
        elif self.align == "top":
            y += self.font_size / 2
        elif self.align == "left":
            x += self.font_size * self.text_width / 2
        elif self.align == "right":
            x -= self.font_size * self.text_width / 2
        elif self.align == "bottomleft":
            y -= self.font_size / 2
            x += self.font_size * self.text_width / 2
        elif self.align == "bottomright":
            y -= self.font_size / 2
            x -= self.font_size * self.text_width / 2
        elif self.align == "topleft":
            y += self.font_size / 2
            x += self.font_size * self.text_width / 2
        elif self.align == "topright":
            y += self.font_size / 2
            x -= self.font_size * self.text_width / 2

        x, y = canvas.to_screen(x, y)

        return canvas.create_window(x, y, window=self.frm)

    def _move(self, dx, dy):
        self.anchor.move(dx, dy)

    def _in_allowed(self, c):
        return c in self.allowed_symbols

    def was_edited(self):
        if self.was_edited_since_last_call:
            self.was_edited_since_last_call = False
            return True
        return False

    def _on_edit(self, *args):
        text = self.text.get()
        if self.max_characters is not None:
            if len(text) > self.max_characters:
                self.text.set(text[:-1])
                return

        corrected = ''.join(filter(self._in_allowed, text))
        self.text.set(corrected)

        if not self.edited:
            self.edited = True
            self.set_text(corrected.replace(self.prompt_text, "", 1))

            if self.text_type == "*":
                self.entry.config(show="*")

        self.was_edited_since_last_call = True

    def _on_enter(self, e):
        pass

    def is_clicked(self, mouse_pos):
        if self.bounds is None:
            if self.entry:
                if mouse_pos is not None:
                    width, height = self.get_width(), self.get_height()
                    if (self.anchor[0] - width / 2 > mouse_pos[0] > self.anchor[0] + width / 2) and \
                       (self.anchor[1] - height / 2 > mouse_pos[1] > self.anchor[1] + height / 2):
                        return True
            return False

        return self.bounds.is_clicked(mouse_pos)

    def allow_character(self, character):
        if character not in self.allowed_symbols and character in ALL_CHARACTERS:
            self.allowed_symbols.append(character)
        return self

    def allow_only_numeric(self, allow=True):
        if allow:
            self.allowed_symbols = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', "-", "."]
        else:
            self.allowed_symbols = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q',
                                    'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                                    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
                                    'Z', '!', '"', '#', '$', '%', '&', '\\', "'", '(', ')', '*', '+', ',', '-', '.',
                                    '/', ':', ';', '<', '=', '>', '?', '@', '[', '\'', '"', ']', '^', '_', '`', '{',
                                    '|', '}', '~', ':', " "]
        return self

    def allow_only_integers(self, allow=True):
        if allow:
            self.allowed_symbols = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', "-"]
        else:
            self.allowed_symbols = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q',
                                    'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                                    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
                                    'Z', '!', '"', '#', '$', '%', '&', '\\', "'", '(', ')', '*', '+', ',', '-', '.',
                                    '/', ':', ';', '<', '=', '>', '?', '@', '[', '\'', '"', ']', '^', '_', '`', '{',
                                    '|', '}', '~', ':', " "]
        return self

    def allow_only_positive_integers(self, allow=True):
        if allow:
            self.allowed_symbols = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        else:
            self.allowed_symbols = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q',
                                    'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                                    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
                                    'Z', '!', '"', '#', '$', '%', '&', '\\', "'", '(', ')', '*', '+', ',', '-', '.',
                                    '/', ':', ';', '<', '=', '>', '?', '@', '[', '\'', '"', ']', '^', '_', '`', '{',
                                    '|', '}', '~', ':', " "]
        return self

    def allow_all_characters(self):
        self.allowed_symbols = ALL_CHARACTERS

    def allow_only_alpha(self, allow=True):
        if allow:
            self.allowed_symbols = self.allowed_symbols = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
                                                           'm','n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
                                                           'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
                                                           'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
                                                           'W', 'X', 'Y', 'Z', " "]
        else:
            self.allowed_symbols = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '!', '"', '#', '$', '%', '&',
                                    '\\', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?',
                                    '@', '[', '\'', '"', ']', '^', '_', '`', '{', '|', '}', '~', ':', " "]
        return self

    def allow_only_custom(self, custom):
        if not set(custom).issubset(set(ALL_CHARACTERS)):
            raise GraphicsError("The Custom list of character contains invalid characters!")

        self.allowed_symbols = custom.copy()
        return self

    def set_max_characters(self, max_characters=None):
        self.max_characters = max_characters

        return self

    def hide_cursor(self):
        self.entry.config(insertontime=0)

    def show_cursor(self):
        self.entry.config(insertontime=1)

    def disable(self):
        self.enabled = "disabled"
        if self.entry:
            self.entry.config(state=self.enabled)

    def enable(self):
        self.enabled = "normal"
        if self.entry:
            self.entry.config(state=self.enabled)

    def toggle_enabled(self):
        self.enabled = ["disabled", "normal"].index(int(self.enabled))
        if self.entry:
            self.entry.config(state=self.enabled)

    def set_enabled(self, active):
        if active not in ["normal", "disabled", "readonly"]:
            raise GraphicsError("\n\nState must be either normal, disabled or readonly")
        self.enabled = active
        if self.entry:
            self.entry.config(state=self.enabled)

    def set_text_type(self, text_type):
        self.text = text_type
        if self.entry:
            self.entry.config(show=self.text)

    def get_text(self):
        return self.text.get()

    def get_anchor(self):
        return self.anchor.clone()

    def get_width(self):
        if self.entry:
            return self.entry.winfo_width()

    def get_height(self):
        if self.entry:
            return self.entry.winfo_height()

    def get_font_size(self):
        return self.font_size

    def get_font_face(self):
        return self.font

    def get_font_colour(self):
        return self.font_colour

    def get_font_style(self):
        return self.font_style

    def get_justify(self):
        return self.justify

    def get_selection_colour(self):
        return self.select_colour

    def get_maximum_characters(self):
        return self.max_characters

    def get_outline_width(self):
        return self.outline_width

    def get_fill(self):
        return self.fill

    def clone(self):
        other = Entry(self.anchor, self.text_width)
        other.config = self.config.copy()
        other.text = tkStringVar()
        other.text.set(self.text.get())
        other.fill = self.fill
        return other

    def set_text(self, t):
        self.text.set(t)
        self._update_layer()
        return self

    def set_fill(self, colour):
        self.fill = colour
        if self.entry:
            self.entry.config(bg=colour)
        self._update_layer()
        return self

    def set_face(self, font_face):
        if font_face in STYLES[self.style].keys():
            self.font = STYLES[self.style][font_face]
        elif isinstance(font_face, str):
            self.font = font_face
        else:
            if "font face" in STYLES[self.style].keys():
                self.font = STYLES[self.style]["font face"]
            else:
                self.font = STYLES["default"]["font face"]

        if self.entry:
            self.entry.config(font=(self.font, self.font_size, self.font_style))
        self._update_layer()
        return self

    def set_font_size(self, font_size, set_initial_font_size=True):
        if isinstance(font_size, int):
            self.font_size = font_size
        elif font_size in STYLES[self.style].keys():
            self.font_size = STYLES[self.style][font_size]
        else:
            if "font size" in STYLES[self.style].keys():
                self.font_size = STYLES[self.style]["font size"]
            else:
                self.font_size = STYLES["default"]["font size"]

        if set_initial_font_size:
            self.initial_font_size = round(font_size)

        if self.entry:
            self.entry.config(font=(self.font, self.font_size, self.font_style))
        self._update_layer()
        return self

    def set_font_style(self, font_style):
        if font_style in STYLES[self.style].keys():
            self.font_style = STYLES[self.style][font_style]
        elif isinstance(font_style, str):
            self.font_style = font_style
        else:
            if "font style" in STYLES[self.style].keys():
                self.font_style = STYLES[self.style]["font style"]
            else:
                self.font_style = STYLES["default"]["font style"]

        if self.entry:
            self.entry.config(font=(self.font, self.font_size, self.font_style))
        self._update_layer()
        return self

    def set_text_colour(self, colour):
        self.font_colour = colour
        if self.entry:
            self.entry.config(fg=colour)
        self._update_layer()
        return self

    def set_border_relief(self, border):
        if border not in ["flat", "groove", "raised", "ridge", "solid", "sunken"]:
            raise GraphicsError("\n\nBorder type must be one of "
                                "['flat', 'groove', 'raised', 'ridge', 'solid', 'sunken']")
        self.border_type = border
        if self.entry:
            self.entry.config(relief=border)
        self._update_layer()
        return self

    def set_width(self, text_width):
        if not isinstance(text_width, int):
            raise GraphicsError(f"Text Width for the Entry must be an integer, not {text_width}")
        self.text_width = text_width

        if self.entry:
            self.entry.config(width=text_width)

        self._update_layer()
        return self



from wave import open as waveopen
from math import sin as mathsin
from struct import pack as structpack

from goopylib.sound.Sound import Sound
from goopylib.util import GraphicsError


def create_sin_wave(filename, duration, freq, tag=None, volume=1, sample_rate=7000):

    if not (isinstance(duration, float) or isinstance(duration, int)):
        raise GraphicsError(f"\n\nGraphicsError: duration value for sin wave must be an int or float, not {duration}")
    if not (isinstance(freq, int) or isinstance(freq, float)):
        raise GraphicsError(f"\n\nGraphicsError: freq value for sin wave must be an int or float, not {freq}")

    if not (isinstance(volume, float) or isinstance(volume, int)):
        raise GraphicsError(f"\n\nGraphicsError: volume for sin wave must be a float or int, not {volume}")
    if not (0 <= volume <= 1):
        raise GraphicsError(f"\n\nGraphicsError: volume for sin wave must be between 0 & 1, not {volume}")

    if not (isinstance(sample_rate, float) or isinstance(sample_rate, int)):
        raise GraphicsError(f"\n\nGraphicsError: sample rate for sin wave must be an integer or float,"
                            f" not {sample_rate}")

    audio = []
    num_samples = int(duration * sample_rate)

    scale = 6.283 * (freq / sample_rate)
    volume *= 127.5
    for x in range(num_samples):
        audio.append(volume * mathsin(x * scale))

    return create_sound_from_values(filename, audio, tag=tag, channels=1, sample_rate=sample_rate)


def create_sawtooth_wave(filename, duration, freq, reverse=False, tag=None, volume=1, sample_rate=7000):
    
    if not (isinstance(duration, float) or isinstance(duration, int)):
        raise GraphicsError(f"\n\nGraphicsError: duration value for sawtooth wave must be an int or float, not {duration}")
    if not (isinstance(volume, float) or isinstance(volume, int)):
        raise GraphicsError(f"\n\nGraphicsError: volume for sawtooth wave must be a float or int, not {volume}")
    if not (0 <= volume <= 1):
        raise GraphicsError(f"\n\nGraphicsError: volume for sawtooth wave must be between 0 & 1, not {volume}")

    if not (isinstance(sample_rate, float) or isinstance(sample_rate, int)):
        raise GraphicsError(f"\n\nGraphicsError: sample rate for sawtooth wave must be an integer or float,"
                            f" not {sample_rate}")
    
    audio = []
    num_samples = int(duration * sample_rate)

    scale = freq / sample_rate
    volume *= 127.5
    for x in range(num_samples):
        x *= scale
        audio.append(volume * (x - int(x)))  # The int function is equivalent to the floor function for positive numbers

    if reverse:
        audio.reverse()

    return create_sound_from_values(filename, audio, tag=tag, channels=1, sample_rate=sample_rate)


def create_square_wave(filename, duration, freq, tag=None, volume=1, sample_rate=7000):
    
    if not (isinstance(duration, float) or isinstance(duration, int)):
        raise GraphicsError(f"\n\nGraphicsError: duration value for square wave must be an int or float, not {duration}")
    if not (isinstance(volume, float) or isinstance(volume, int)):
        raise GraphicsError(f"\n\nGraphicsError: volume for square wave must be a float or int, not {volume}")
    if not (0 <= volume <= 1):
        raise GraphicsError(f"\n\nGraphicsError: volume for square wave must be between 0 & 1, not {volume}")

    if not (isinstance(sample_rate, float) or isinstance(sample_rate, int)):
        raise GraphicsError(f"\n\nGraphicsError: sample rate for square wave must be an integer or float,"
                            f" not {sample_rate}")
    
    audio = [0]
    num_samples = duration * sample_rate

    volume *= 127.5

    i = 0
    f = 0

    check = sample_rate / (freq * 0.5)
    for _ in range(num_samples):
        if f > check:
            i = 1 - i
            f = 0
            audio.append(volume * i)
        else:
            f += 1
            audio.append(audio[-1])
    audio.pop()

    return create_sound_from_values(filename, audio, tag=tag, channels=1, sample_rate=sample_rate)


def create_sound_from_values(filename, values, tag=None, channels=1, sample_rate=11050):
    
    if not (tag is None or isinstance(tag, str)):
        raise GraphicsError(f"\n\nGraphicsError: tag for sound must be a string or None, not {tag}")
    if not (isinstance(channels, int)):
        raise GraphicsError(f"\n\nGraphicsError: number of channels for audio must be an integer, not {channels}")
    if not (isinstance(sample_rate, float) or isinstance(sample_rate, int)):
        raise GraphicsError(f"\n\nGraphicsError: sample rate for audio must be an integer or float,"
                            f" not {sample_rate}")

    with waveopen(filename, "w") as wav_file:
        sampwidth = 1

        nframes = len(values)
        comptype = "NONE"

        wav_file.setparams((channels, sampwidth, sample_rate, nframes, comptype, "not compressed"))

        minimum = min(values)
        if minimum < 0:
            for i in range(len(values)):
                values[i] -= minimum

        for sample in values:
            wav_file.writeframes(structpack('B', int(sample)))

    return Sound(filename, tag=filename if tag is None else tag)
